from itertools import count
from typing import Iterator

from .objects import Result, ResultWithDetails
from .session import session

__all__ = ["Results", "ResultDetail"]


RESULTS_PER_PAGE = 50


class Results:
    """
    Interfaces with the evaluations of smartschool.

    To reproduce: "Ga naar" > "Resultaten", it'll be one of the XHR calls then.

    Example:
    -------
    >>> for result in Results():
    >>>     print(result.name)
    Repetitie hoofdstuk 1

    """

    def __iter__(self) -> Iterator[Result]:
        for page_nr in count(start=1):  # pragma: no branch
            json = session.json(f"/results/api/v1/evaluations/?pageNumber={page_nr}&itemsOnPage={RESULTS_PER_PAGE}")
            for result in json:
                yield Result(**result)

            if len(json) < RESULTS_PER_PAGE:
                break


class ResultDetail:
    def __init__(self, result_id: str):
        self.result_id = result_id

    def get(self) -> ResultWithDetails:
        json = session.json(f"/results/api/v1/evaluations/{self.result_id}")
        return ResultWithDetails(**json)



from __future__ import annotations

import os
from abc import ABC
from dataclasses import dataclass, field
from pathlib import Path

import yaml


class Credentials(ABC):
    username: str
    password: str
    main_url: str

    other_info: dict | None = None

    def validate(self) -> None:
        self.username = (self.username or "").strip()
        self.password = (self.password or "").strip()
        self.main_url = (self.main_url or "").strip()

        error = []
        if not self.username:
            error.append("username")
        if not self.password:
            error.append("password")
        if not self.main_url:
            error.append("main_url")

        if error:
            raise RuntimeError(f"Please verify and correct these attributes: {error}")


@dataclass
class PathCredentials(Credentials):
    filename: str | Path = field(default=Path.cwd().joinpath("credentials.yml"))

    def __post_init__(self):
        self.filename = Path(self.filename)

        cred_file: dict = yaml.safe_load(self.filename.read_text(encoding="utf8"))
        self.username = cred_file.pop("username", None)
        self.password = cred_file.pop("password", None)
        self.main_url = cred_file.pop("main_url", None)

        self.other_info = cred_file


@dataclass
class EnvCredentials(Credentials):
    def __post_init__(self):
        self.username = os.getenv("SMARTSCHOOL_USERNAME")
        self.password = os.getenv("SMARTSCHOOL_PASSWORD")
        self.main_url = os.getenv("SMARTSCHOOL_MAIN_URL")



from __future__ import annotations

import base64
from datetime import date, datetime
from functools import cached_property
from typing import Annotated, Literal

from pydantic import AliasChoices, BeforeValidator, constr
from pydantic.dataclasses import Field, dataclass

from .common import as_float
from .session import session

String = constr(strip_whitespace=True)


def convert_to_datetime(x: str | String | datetime) -> datetime:
    if isinstance(x, datetime):
        if x.tzinfo is None:
            raise ValueError("No timezone information found in this date")
        return x

    try:
        return datetime.strptime(x, "%Y-%m-%dT%H:%M:%S%z")
    except ValueError:  # 2023-11-16 08:24
        return datetime.strptime(x, "%Y-%m-%d %H:%M")


def convert_to_date(x: str | String | date | datetime) -> date:
    if isinstance(x, datetime):
        return x.date()
    if isinstance(x, date):
        return x

    return datetime.strptime(x, "%Y-%m-%d").date()


Url = Annotated[str | String, BeforeValidator(lambda x: session.create_url(x))]
Date = Annotated[date, BeforeValidator(convert_to_date)]
DateTime = Annotated[datetime, BeforeValidator(convert_to_datetime)]


@dataclass
class CourseGraphic:
    type: Literal["icon"]
    value: String


@dataclass
class ResultGraphic:
    type: Literal["percentage"]
    color: Literal["green", "red"]
    value: int
    description: String

    @cached_property
    def achieved_points(self) -> float:
        return as_float(self.description.split("/")[0])

    @cached_property
    def total_points(self) -> float:
        return as_float(self.description.split("/")[1])

    @property
    def percentage(self) -> float:
        return self.achieved_points / self.total_points


@dataclass
class PersonDescription:
    startingWithFirstName: String = ""
    startingWithLastName: String = ""


@dataclass
class _User:
    id: String
    pictureHash: String
    pictureUrl: Url
    description: PersonDescription
    name: PersonDescription
    sort: String


@dataclass
class DateRange:
    start: DateTime
    end: DateTime


@dataclass
class SkoreWorkYear:
    id: int
    dateRange: DateRange


@dataclass
class Class_:
    identifier: String
    id: int
    platformId: int
    name: String
    type: String
    icon: String


@dataclass
class Period:
    id: int
    name: String
    icon: String
    skoreWorkYear: SkoreWorkYear
    isActive: bool
    class_: Class_ = Field(validation_alias=AliasChoices("class", "class_"))


@dataclass
class Component:
    id: int
    name: String
    abbreviation: String


Teacher = _User
Student = _User


@dataclass
class Course:
    id: int
    name: String
    graphic: CourseGraphic
    teachers: list[Teacher]
    skoreClassId: int
    parentCourseId: None | int
    skoreWorkYear: SkoreWorkYear
    class_: Class_ = Field(validation_alias=AliasChoices("class", "class_"))


@dataclass
class Feedback:
    text: String
    user: Teacher


@dataclass
class FeedbackFull:
    attachments: list[String]
    changedAt: DateTime
    createdAt: DateTime
    evaluationId: String
    id: String
    student: Student
    teacher: Teacher
    text: String


@dataclass
class Result:
    identifier: String
    type: Literal["normal"]
    name: String
    graphic: ResultGraphic
    date: DateTime

    gradebookOwner: Teacher
    component: Component
    courses: list[Course]
    period: Period
    feedback: list[Feedback]
    feedbacks: list[FeedbackFull]
    availabilityDate: DateTime
    isPublished: bool
    doesCount: bool


@dataclass
class ResultDetails:
    centralTendencies: list[String]
    teachers: list[Teacher]
    dateChanged: DateTime
    userChanged: Teacher
    class_: Class_ = Field(validation_alias=AliasChoices("class", "class_"))


@dataclass
class ResultWithDetails(Result):
    details: ResultDetails


@dataclass
class CourseCondensed:
    name: String
    teacher: String
    url: Url

    descr: String = Field(repr=False, default="")
    icon: String = Field(repr=False, default="")


@dataclass
class FutureTaskOneTask:
    label: String
    description: String
    icon: String
    warning: bool
    click_handle: String
    activityID: int
    dateID: String
    assignmentID: String
    endMomentTS: String | None
    startMomentID: String
    endMomentID: String
    lessonID: String
    type: String
    classID: String
    course: String
    date: Date
    hourID: String


@dataclass
class FutureTaskOneItem:
    tasks: list[FutureTaskOneTask]
    materials: list[String]


@dataclass
class FutureTaskOneCourse:
    lessonID: String
    hourID: String
    classID: String
    course_title: String
    items: FutureTaskOneItem


@dataclass
class FutureTaskOneDay:
    date: Date
    pretty_date: String
    courses: list[FutureTaskOneCourse]


@dataclass
class FutureTasks:
    """
    Class that interfaces the retrieval of any task that needs to be made in the near future.

    Example:
    -------
    >>> for day in FutureTasks().days:
    >>>     for course in day.courses:
    >>>         print("Course:", course.course_title)
    >>>         for task in course.items.tasks:
    >>>             print("Task:", task.description)
    Course: 2 - AAR1, Lotte Peeters
    Task: Toets 3. De koolstofcyclus in het systeem aarde pagina 42 - 47

    """

    days: list[FutureTaskOneDay] = Field(default_factory=list)
    last_assignment_id: int = 0
    last_date: Date = Field(default_factory=date.today)

    def __post_init__(self):
        """I need to do this here because when I do it in Agenda, it'll not lazily load it. But in this way, I load it on construction."""
        json = session.json(
            "/Agenda/Futuretasks/getFuturetasks",
            method="post",
            data={
                "lastAssignmentID": 0,
                "lastDate": "",
                "filterType": "false",
                "filterID": "false",
            },
            headers={
                "X-Requested-With": "XMLHttpRequest",
            },
        )

        self.days = []
        for d in json["days"]:
            self.days.append(FutureTaskOneDay(**d))

        self.last_assignment_id = json["last_assignment_id"]
        self.last_date = convert_to_date(json["last_date"])


@dataclass
class AgendaHour:
    hourID: String
    start: String
    end: String
    title: String


@dataclass
class AgendaLesson:
    momentID: String
    lessonID: String
    hourID: String
    date: Date
    subject: String | None
    course: String
    courseTitle: String
    classroom: String
    classroomTitle: String
    teacher: String
    teacherTitle: String
    klassen: String
    klassenTitle: String
    classIDs: String
    bothStartStatus: String
    assignmentEndStatus: String
    testDeadlineStatus: String
    noteStatus: String
    note: String | None
    date_listview: String
    hour: String
    activity: String
    activityID: String | None
    color: String
    hourValue: String
    components_hidden: object
    freedayIcon: String
    someSubjectsEmpty: String | None

    @property
    def hour_details(self) -> AgendaHour:
        from .agenda import SmartschoolHours

        return SmartschoolHours().search_by_hourId(self.hourID)


@dataclass
class AgendaMomentInfoAssignment:
    startAssignment: String
    start: String
    end: String
    type: String
    description: String
    atdescription: String
    freedeadline: String
    warning: String
    assignmentInfo: String
    assignmentDeadline: String


@dataclass
class AgendaMomentInfo:
    className: String
    subject: String
    materials: String | None
    momentID: String
    assignments: list[AgendaMomentInfoAssignment]


@dataclass
class StudentSupportLink:
    id: String
    name: String
    description: String
    icon: String
    link: Url
    cleanLink: String
    isVisible: bool


@dataclass
class ShortMessage:
    id: int
    fromImage: Url
    subject: String
    date: DateTime
    status: int
    attachment: int
    unread: bool
    label: bool
    deleted: bool
    allowreply: bool
    allowreplyenabled: bool
    hasreply: bool
    hasForward: bool
    realBox: String
    sendDate: DateTime | None
    from_: String = Field(validation_alias=AliasChoices("from", "from_"))


@dataclass
class FullMessage:
    id: int
    to: String | None
    subject: String
    date: DateTime
    body: String
    status: int
    attachment: int
    unread: bool
    label: bool
    receivers: list[String]
    ccreceivers: list[String]
    bccreceivers: list[String]
    senderPicture: String
    markedInLVS: None
    fromTeam: int
    totalNrOtherToReciviers: int
    totalnrOtherCcReceivers: int
    totalnrOtherBccReceivers: int
    canReply: bool
    hasReply: bool
    hasForward: bool
    sendDate: DateTime | None
    from_: String = Field(validation_alias=AliasChoices("from", "from_"))


@dataclass
class Attachment:
    fileID: int
    name: String
    mime: String
    size: String
    icon: String
    wopiAllowed: bool
    order: int

    def download(self) -> bytes:
        resp = session.get(f"/?module=Messages&file=download&fileID={self.fileID}&target=0")
        return base64.b64decode(resp.content)


@dataclass
class MessageChanged:
    id: int
    new: int = Field(validation_alias=AliasChoices("status", "label", "new"))


@dataclass
class MessageDeletionStatus:
    msgID: int
    boxType: String
    is_deleted: bool = Field(validation_alias=AliasChoices("status", "is_deleted"))



from __future__ import annotations

from functools import cached_property
from typing import Iterator

from .objects import Course, CourseCondensed
from .session import session

__all__ = ["Courses", "TopNavCourses"]


class TopNavCourses:
    """
    Retrieves a list of the courses which are available from the top navigation bar.

    This structure is different from the `Courses` results.

    Example:
    -------
    >>> for course in TopNavCourses():
    >>>     print(course.name)
    Aardrijkskunde_3_LOP_2023-2024
    bibliotheek

    """

    @cached_property
    def _list(self) -> list[CourseCondensed]:
        return [CourseCondensed(**course) for course in session.json("/Topnav/getCourseConfig", method="post")["own"]]

    def __iter__(self) -> Iterator[CourseCondensed]:
        yield from self._list


class Courses:
    """
    Retrieves a list of the courses.

    This structure is different from the `TopNavCourses` results.

    To reproduce: go to "Results", one of the XHR calls is this one

    Example:
    -------
    >>> for course in Courses():
    >>>     print(course.name)
    Aardrijkskunde
    Biologie

    """

    @cached_property
    def _list(self) -> list[Course]:
        return [Course(**course) for course in session.json("/results/api/v1/courses/")]

    def __iter__(self) -> Iterator[Course]:
        yield from self._list



from __future__ import annotations

import time
from abc import ABC
from datetime import datetime

from ._xml_interface import SmartschoolXML_WeeklyCache
from .objects import AgendaHour, AgendaLesson, AgendaMomentInfo


class AgendaPoster(SmartschoolXML_WeeklyCache, ABC):
    """Caches the information on a weekly basis, and posts to the mentioned URL."""

    _url: str = "/?module=Agenda&file=dispatcher"


class SmartschoolLessons(AgendaPoster):
    """
    Interface to the retrieval of lessons for a certain date.

    To reproduce: open the agenda, one of the XHR calls is this one.

    This includes (for an example check the tests):
    - momentID: a unique identifier for this specific lesson
    - lessonID: a group id
    - hourID: link to `AgendaHours`
    - date: YYYY-mm-dd
    - subject: explanation by the teacher
    - course
    - courseTitle
    - classroom: what classroom this is in
    - classroomTitle: same as `classroom`?
    - teacher: name of the teacher (Lastname + initial of firstname)
    - teacherTitle:
    - klassen
    - klassenTitle
    - classIDs
    - bothStartStatus
    - assignmentEndStatus
    - testDeadlineStatus
    - noteStatus
    - note
    - date_listview
    - hour
    - activity
    - activityID
    - color
    - hourValue
    - components_hidden
    - freedayIcon
    - someSubjectsEmpty
    """

    @property
    def _xpath(self) -> str:
        return ".//lesson"

    @property
    def _object_to_instantiate(self) -> type[AgendaLesson]:
        return AgendaLesson

    @property
    def _subsystem(self) -> str:
        return "agenda"

    @property
    def _action(self) -> str:
        return "get lessons"

    @property
    def _params(self) -> dict:
        now = (self.timestamp_to_use or datetime.now()).timestamp()
        in_5_days = now + 5 * 24 * 3600

        return {
            "startDateTimestamp": now,  # 1700045313
            "endDateTimestamp": in_5_days,  # 1700477313
            "filterType": "false",
            "filterID": "false",
            "gridType": "1",
            "classID": "0",
            "endDateTimestampOld": in_5_days,  # 1700477313
            "forcedTeacher": "0",
            "forcedClass": "0",
            "forcedClassroom": "0",
            "assignmentTypeID": "1",
        }


class SmartschoolHours(AgendaPoster):
    """
    Interface to the retrieval of periods (called Hours in smartschool).

    To reproduce: open the agenda, one of the XHR calls is this one.

    This includes (for an example check the tests):
    - hourID: unique identifier for this period
    - start: HH:MM starting time
    - end: HH:MM ending time
    - title: how it is called in the agenda
    """

    @property
    def _xpath(self) -> str:
        return ".//hour"

    @property
    def _object_to_instantiate(self) -> type[AgendaHour]:
        return AgendaHour

    @property
    def _subsystem(self) -> str:
        return "grid"

    @property
    def _action(self) -> str:
        return "get hours"

    @property
    def _params(self) -> dict:
        return {"date": int(time.time())}

    def search_by_hourId(self, hourId: str):
        for hour in self._xml():
            if hour.hourID == hourId:
                return hour

        raise ValueError(f"Couldn't find {hourId}")


class SmartschoolMomentInfos(AgendaPoster):
    """
    Interface to the retrieval of one particular moment (a book-symbol in smartschool).

    To reproduce: open the agenda, click on a yellow/red book. This is the XHR call that appears.

    This includes (for an example check the tests):
    - hourID: unique identifier for this period
    - start: HH:MM starting time
    - end: HH:MM ending time
    - title: how it is called in the agenda
    """

    def __init__(self, moment_id: str):
        super().__init__()

        moment_id = str(moment_id).strip()
        if not moment_id:
            raise ValueError("Please provide a valid MomentID")

        self._moment_id = moment_id

    @property
    def _xpath(self) -> str:
        return ".//class"

    @property
    def _object_to_instantiate(self) -> type[AgendaMomentInfo]:
        return AgendaMomentInfo

    @property
    def _subsystem(self) -> str:
        return "agenda"

    @property
    def _action(self) -> str:
        return "get moment info"

    @property
    def _params(self) -> dict:
        return {
            "momentID": self._moment_id,
            "dateID": "",
            "assignmentIDs": "",
            "activityID": "0",
        }

    def _post_process_element(self, element: dict) -> None:
        """In this XML we have `assignments.assignment`, but I don't need that `assignment` tag."""
        if element["assignments"] is None:
            element["assignments"] = []
            return

        if isinstance(element["assignments"]["assignment"], list):
            element["assignments"] = element["assignments"]["assignment"]
        else:
            element["assignments"] = [element["assignments"]["assignment"]]



from __future__ import annotations

import contextlib
import functools
import json
import time
from dataclasses import dataclass, field
from functools import cached_property
from http.cookiejar import LWPCookieJar
from pathlib import Path
from typing import TYPE_CHECKING, Self
from urllib.parse import urljoin

from requests import Session

from .common import bs4_html, get_all_values_from_form

if TYPE_CHECKING:  # pragma: no cover
    from requests import Response

    from .credentials import Credentials


def _handle_cookies_and_login(func):
    @functools.wraps(func)
    def inner(self: Smartschool, *args, **kwargs):
        if self.creds is None:
            raise RuntimeError("Please start smartschool first via: `Smartschool.start(PathCredentials())`")

        self._try_login()

        resp = func(self, *args, **kwargs)

        self._session.cookies.save(ignore_discard=True)

        return resp

    return inner


@dataclass
class Smartschool:
    creds: Credentials = None

    already_logged_on: bool = field(init=False, default=None)
    _session: Session = field(init=False, default_factory=Session)

    def _try_login(self) -> None:
        if self.already_logged_on is None:
            # Created in the last 10 minutes? Assume we're still logged on...
            self.already_logged_on = self.cookie_file.exists() and self.cookie_file.stat().st_mtime > (time.time() - 600)

        if self.already_logged_on:
            return

        self.already_logged_on = True

        resp = session.get("/login")  # This will either log you in, or redirect to the main page. Refreshing the cookies in the meanwhile
        if resp.url.endswith("/login"):  # Not redirect >> do log in
            session._do_login(resp)

    @classmethod
    def start(cls, creds: Credentials) -> Self:
        global session

        creds.validate()
        session.creds = creds

        return session

    @property
    def cookie_file(self) -> Path:
        return Path.cwd() / "cookies.txt"

    def __post_init__(self):
        self._session.headers["User-Agent"] = "unofficial Smartschool API interface"

        cookie_jar = LWPCookieJar(self.cookie_file)
        with contextlib.suppress(FileNotFoundError):
            cookie_jar.load(ignore_discard=True)

        self._session.cookies = cookie_jar

    def create_url(self, url: str) -> str:
        return urljoin(self._url, url)

    @_handle_cookies_and_login
    def post(self, url, *args, **kwargs) -> Response:
        return self._session.post(self.create_url(url), *args, **kwargs)

    @_handle_cookies_and_login
    def get(self, url, *args, **kwargs) -> Response:
        return self._session.get(self.create_url(url), *args, **kwargs)

    def json(self, url, *args, method: str = "get", **kwargs) -> dict:
        """Sometimes this is double json-encoded. So we keep trying until it isn't a string anymore."""
        if method.lower() == "post":
            r = self.post(url, *args, **kwargs)
        else:
            r = self.get(url, *args, **kwargs)

        json_ = r.text

        while isinstance(json_, str):
            json_ = json.loads(json_)

        return json_

    def _do_login(self, response: Response) -> Response:
        html = bs4_html(response)
        inputs = get_all_values_from_form(html, 'form[name="login_form"]')

        final = 0
        data = {}
        for input_ in inputs:
            if "username" in input_["name"]:
                data[input_["name"]] = self.creds.username
                final |= 1
            elif "password" in input_["name"]:
                data[input_["name"]] = self.creds.password
                final |= 2
            else:
                data[input_["name"]] = input_["value"]

        assert final == 3, "We're missing something here!"

        return self.post(response.url, data=data)

    @cached_property
    def _url(self) -> str:
        return "https://" + self.creds.main_url

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(for: {self.creds.username})"


session: Smartschool = Smartschool()



import logging

from .agenda import SmartschoolHours, SmartschoolLessons, SmartschoolMomentInfos
from .courses import Courses, TopNavCourses
from .credentials import EnvCredentials, PathCredentials
from .logger import setup_logger
from .messages import (
    AdjustMessageLabel,
    Attachments,
    BoxType,
    MarkMessageUnread,
    Message,
    MessageHeaders,
    MessageLabel,
    MessageMoveToArchive,
    MessageMoveToTrash,
    SortField,
    SortOrder,
)
from .objects import FutureTasks
from .periods import Periods
from .results import ResultDetail, Results
from .session import Smartschool
from .student_support import StudentSupportLinks

__all__ = [
    "PathCredentials",
    "EnvCredentials",
    "Smartschool",
    "logger",
    "Courses",
    "TopNavCourses",
    "Results",
    "Periods",
    "FutureTasks",
    "SortField",
    "SortOrder",
    "BoxType",
    "MessageHeaders",
    "StudentSupportLinks",
    "SmartschoolHours",
    "SmartschoolLessons",
    "SmartschoolMomentInfos",
    "Message",
    "Attachments",
    "MarkMessageUnread",
    "AdjustMessageLabel",
    "MessageMoveToArchive",
    "MessageMoveToTrash",
    "MessageLabel",
    "ResultDetail",
]

logger = setup_logger(logging.DEBUG)



from __future__ import annotations

import contextlib
import functools
import inspect
import json
import operator
import platform
import re
import smtplib
import sys
import traceback
import warnings
import xml.etree.ElementTree as ET
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from enum import Enum, auto
from pathlib import Path
from typing import Any, Callable, Literal

from bs4 import BeautifulSoup, FeatureNotFound, GuessedAtParserWarning
from pydantic import RootModel
from pydantic.dataclasses import is_pydantic_dataclass
from requests import Response

__all__ = [
    "send_email",
    "capture_and_email_all_exceptions",
    "save",
    "IsSaved",
    "bs4_html",
    "get_all_values_from_form",
    "make_filesystem_safe",
    "as_float",
    "xml_to_dict",
]

_used_bs4_option = None


class IsSaved(Enum):
    NEW = auto()
    UPDATED = auto()
    SAME = auto()


def save(
    type_: Literal["agenda", "punten", "todo"], course_name: str, id_: str, data: dict | str | Any, is_eq: Callable = operator.eq, extension: str = "json"
) -> IsSaved | dict | str:
    save_as = Path.cwd() / f".cache/_{type_}/{course_name}/{id_}.{extension}"
    save_as.parent.mkdir(exist_ok=True, parents=True)
    data_was_dict = isinstance(data, dict)
    data_was_object = is_pydantic_dataclass(data.__class__)

    if data_was_dict:
        to_write = json.dumps(data, indent=4)
    elif data_was_object:
        to_write = RootModel[data.__class__](data).model_dump_json(indent=4)
    else:
        to_write = data

    if not save_as.exists():
        save_as.write_text(to_write, encoding="utf8")
        return IsSaved.NEW

    old_data = save_as.read_text(encoding="utf8")
    if data_was_dict or data_was_object:
        old_data = json.loads(old_data)
    if data_was_object:
        old_data = data.__class__(**old_data)

    if is_eq(data, old_data):
        return IsSaved.SAME

    save_as.write_text(to_write, encoding="utf8")
    return old_data


def send_email(
    subject: str,
    text: str,
    email_to: list[str] | str,
    email_from: str,
):
    if isinstance(email_to, str):
        email_to = [email_to]

    message = MIMEMultipart("alternative")
    message["Subject"] = subject
    message["From"] = email_from
    message["To"] = ", ".join(email_to)
    message.attach(MIMEText(text, "plain", "utf8"))

    print(f"Sending email >> {subject}")

    if platform.system() == "Windows":
        print("=================== On Linux we would have sent this: ===================")
        print(f"Subject: {subject}")
        print("")
        print(text)
        print("=========================================================================")
        return

    with smtplib.SMTP("localhost") as server:
        server.sendmail(
            from_addr=email_from,
            to_addrs=email_to,
            msg=message.as_string(),
        )


def capture_and_email_all_exceptions(
    email_from: str | list[str], email_to: str | list[str], subject: str = "[⚠Smartschool parser⚠] Something went wrong"
) -> Callable:
    def decorator(func):
        @functools.wraps(func)
        def inner(*args, **kwargs):
            frm = inspect.stack()[1]
            module_name = Path(frm.filename)
            function_signature = f"{module_name.stem}.{func.__name__}"

            print(f"[{function_signature}] Start")
            try:
                result = func(*args, **kwargs)
            except Exception as ex:
                print(f"[{function_signature}] An exception happened: {ex}")

                send_email(
                    email_to=email_to,
                    email_from=email_from,
                    subject=subject,
                    text="".join(traceback.format_exception(None, ex, ex.__traceback__)),
                )

                sys.exit(1)

            print(f"[{function_signature}] Finished")
            return result

        return inner

    return decorator


def bs4_html(html: str | bytes | Response) -> BeautifulSoup:
    global _used_bs4_option

    if isinstance(html, Response):
        html = html.text

    possible_options = [
        _used_bs4_option,
        {"parser": "html.parser", "features": "lxml"},
        {"features": "html5lib"},
        {"features": "html.parser"},
    ]

    for kw in possible_options:  # pragma: no branch
        if kw is None:
            continue

        with contextlib.suppress(FeatureNotFound):
            parsed = BeautifulSoup(html, **kw)
            _used_bs4_option = kw
            return parsed

    _used_bs4_option = {}
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=GuessedAtParserWarning)
        return BeautifulSoup(html)


def get_all_values_from_form(html, form_selector):
    form = html.select(form_selector)
    assert len(form) == 1, f"We should have only 1 form. We got {len(form)}!"
    form = form[0]

    # action = form.attrs.get("action").lower()
    # method = form.attrs.get("method", "get").lower()
    all_inputs = form.find_all(["input", "button", "textarea", "select"])

    inputs = []
    for input_tag in all_inputs:
        tag_name = input_tag.name.lower()
        attrs = input_tag.attrs

        if "name" not in attrs:
            continue

        assert tag_name != "select", "Check if this code works. Possible issue with getting the value if the value tag isn't set"
        inputs.append(
            {
                "name": attrs.get("name"),
                "value": attrs.get("value", ""),
            }
        )
        #     select_options = []
        #     value = ""
        #     for select_option in input_tag.find_all("option"):
        #         option_value = select_option.attrs.get("value")
        #         if option_value:
        #             select_options.append(option_value)
        #             if "selected" in select_option.attrs:
        #                 value = option_value
        #     if not value and select_options:
        #         # if the default is not set, and there are options, take the first option as default
        #         value = select_options[0]
        #     # add the select to the inputs list
        #     inputs.append({"name": attrs.get("name"), "values": select_options, "value": value})

    return inputs


def make_filesystem_safe(name: str) -> str:
    name = re.sub("[^-_a-z0-9.]+", "_", name, flags=re.IGNORECASE)
    name = re.sub("_{2,}", "_", name)
    return name


def as_float(txt: str) -> float:
    return float(txt.replace(",", "."))


def xml_to_dict(element, *, depth: int = 0):
    if depth == 0 and isinstance(element, str):
        element = ET.XML(element.strip())

    result = {}

    for child in element:
        tag = child.tag
        if len(child) == 0:  # Leaf node?
            child_data = child.text
        else:
            child_data = xml_to_dict(child)

        if tag in result:
            if not isinstance(result[tag], list):
                # If the tag already exists, convert it to a list
                result[tag] = [result[tag]]
            result[tag].append(child_data)
        else:
            result[tag] = child_data

    return result



import logging


def setup_logger(lowest_level: int = logging.INFO) -> logging.Logger:
    default_format = "[%(asctime)s] [%(levelname)s] %(name)s > %(message)s"
    logging.basicConfig(format=default_format, level=lowest_level)

    formatter = logging.Formatter(default_format)

    logger = logging.getLogger("smartschool")
    for h in logging.root.handlers:
        h.setFormatter(formatter)

    logger.setLevel(lowest_level)

    return logger



from typing import Iterator

from .objects import Period
from .session import session

__all__ = ["Periods"]


class Periods:
    """
    Retrieves a list of the periods.

    To reproduce: go to "Results", one of the XHR calls is this one

    Example:
    -------
    >>> for period in Periods():
    >>>     print(period.name)
    1 september - 24 oktober
    25 oktober - 19 december

    """

    def __iter__(self) -> Iterator[Period]:
        for period in session.json("/results/api/v1/periods/"):
            yield Period(**period)



from typing import Iterator

from .objects import StudentSupportLink
from .session import session

__all__ = ["StudentSupportLinks"]


class StudentSupportLinks:
    """
    Interfaces with the link section that is loaded when opening the main interface of smartschool.

    To reproduce: Just open the homepage, it'll be one of the XHR calls then.

    Example:
    -------
    >>> for link in StudentSupportLinks():
    >>>     print(link.name)
    1712
    Autisme chat

    """

    def __iter__(self) -> Iterator[StudentSupportLink]:
        json = session.json("/student-support/api/v1/")
        for result in json:
            yield StudentSupportLink(**result)



from __future__ import annotations

import contextlib
from abc import ABC, ABCMeta, abstractmethod
from datetime import date
from typing import TYPE_CHECKING, Iterator, TypeVar
from xml.etree import ElementTree as ET
from xml.sax.saxutils import quoteattr

from .common import xml_to_dict
from .session import session

if TYPE_CHECKING:  # pragma: no cover
    from datetime import datetime

_T = TypeVar("_T")


class _SmartschoolXMLMeta(ABCMeta):
    """
    The metaclass will ONLY create this cache dictionary ones per class instantiation.

    Effectively giving each derived class I make their own dictionary.
    """

    def __new__(cls, name, bases, dct):
        dct["cache"] = {}
        return super().__new__(cls, name, bases, dct)


class SmartschoolXML(ABC, metaclass=_SmartschoolXMLMeta):
    cache: dict  # Type hint for the dynamically added attribute

    def _construct_command(self) -> str:
        txt = "<request><command>"
        txt += f"<subsystem>{self._subsystem}</subsystem>"
        txt += f"<action>{self._action}</action>"
        txt += "<params>"

        for k, v in self._params.items():
            txt += f"<param name={quoteattr(k)}><![CDATA[{v}]]></param>"

        txt += "</params></command></request>"
        return txt

    def __iter__(self) -> Iterator[_T]:
        yield from self._xml()

    def get(self) -> _T:
        """
        Retrieve only the first entry.

        This is particularly useful when you want to fetch just one entry, and don't want to use the `list(Class())[0]` syntax.
        Instead, you can use `Class().get()`.
        """
        return next(iter(self))

    @abstractmethod
    def _get_from_cache(self) -> object:
        """
        Retrieves information from the cache.

        Throws a `KeyError` when not found.
        """

    @abstractmethod
    def _store_into_cache(self, obj: object) -> None:
        """Stores the current value into the cache."""

    @property
    @abstractmethod
    def _url(self) -> str:
        """The URL where the data should be posted to."""

    def _xml(self):
        with contextlib.suppress(KeyError):
            return self._get_from_cache()

        response = session.post(
            self._url,
            data={
                "command": self._construct_command(),
            },
            headers={
                "X-Requested-With": "XMLHttpRequest",
            },
        )

        root = ET.fromstring(response.text)

        all_entries = []
        as_obj = self._object_to_instantiate
        for el in root.findall(self._xpath):
            as_dict = xml_to_dict(el)
            self._post_process_element(as_dict)
            obj = as_obj(**as_dict)
            all_entries.append(obj)

        self._store_into_cache(all_entries)

        return all_entries

    @property
    @abstractmethod
    def _subsystem(self) -> str:
        """Returns the subsystem to request the info from."""

    @property
    @abstractmethod
    def _action(self) -> str:
        """Returns the action to send."""

    @property
    @abstractmethod
    def _params(self) -> dict:
        """Returns the parameters to send."""

    @property
    @abstractmethod
    def _xpath(self) -> str:
        """Returns the xpath to investigate."""

    @property
    @abstractmethod
    def _object_to_instantiate(self) -> type[_T]:
        """Returns the object to instantiate."""

    def _post_process_element(self, element: dict) -> None:
        """By default, this doesn't do anything, but you can adjust the parsed XML when needed."""


class SmartschoolXML_WeeklyCache(SmartschoolXML, ABC):
    @property
    def _cache_key(self):
        # Week number
        today = self.timestamp_to_use or date.today()
        return today.strftime("%Y-%U")

    def _get_from_cache(self) -> object:
        return self.cache[self._cache_key]

    def _store_into_cache(self, obj: object) -> None:
        self.cache[self._cache_key] = obj

    def __init__(self, timestamp_to_use: datetime | None = None):
        self.timestamp_to_use = timestamp_to_use


class SmartschoolXML_NoCache(SmartschoolXML, ABC):
    def _get_from_cache(self) -> object:
        raise KeyError

    def _store_into_cache(self, _: object) -> None:
        return



from __future__ import annotations

from abc import ABC
from enum import Enum
from typing import Iterator
from urllib.parse import quote_plus

from ._xml_interface import SmartschoolXML, SmartschoolXML_NoCache
from .objects import Attachment, FullMessage, MessageChanged, MessageDeletionStatus, ShortMessage
from .session import session

__all__ = [
    "SortField",
    "SortOrder",
    "BoxType",
    "MessageHeaders",
    "Message",
    "Attachments",
    "MarkMessageUnread",
    "AdjustMessageLabel",
    "MessageMoveToArchive",
    "MessageMoveToTrash",
    "MessageLabel",
]


class SortField(Enum):
    DATE = "date"
    FROM = "from"
    READ_UNREAD = "status"
    ATTACHMENT = "attachment"
    FLAG = "label"


class SortOrder(Enum):
    ASC = "asc"
    DESC = "desc"


class BoxType(Enum):
    INBOX = "inbox"
    DRAFT = "draft"
    SCHEDULED = "scheduled"
    SENT = "outbox"
    TRASH = "trash"


class MessageLabel(Enum):
    NO_FLAG = 0
    GREEN_FLAG = 1
    YELLOW_FLAG = 2
    RED_FLAG = 3
    BLUE_FLAG = 4


class _MessagesPoster:
    _url = "/?module=Messages&file=dispatcher"


class MessageHeaders(_MessagesPoster, SmartschoolXML_NoCache):
    """
    Interfaces the mailbox principle in Smartschool.

    To reproduce: Open the messages and watch the XHR calls.

    Example:
    -------
    >>> for message in MessageHeaders():
    >>>     print(message.subject)
    Re: LO les
    Frans

    """

    def __init__(
        self,
        box_type: BoxType = BoxType.INBOX,
        sort_by: SortField = SortField.DATE,
        sort_order: SortOrder = SortOrder.DESC,
        already_seen_message_ids: list[int] | None = None,
    ):
        super().__init__()

        self.box_type = box_type
        self.sort_by = sort_by
        self.sort_order = sort_order
        self.already_seen_message_ids = already_seen_message_ids or []

    @property
    def _subsystem(self) -> str:
        return "postboxes"

    @property
    def _action(self) -> str:
        return "message list"

    @property
    def _params(self) -> dict:
        return {
            "boxType": self.box_type.value,
            "boxID": "0",
            "sortField": self.sort_by.value,
            "sortKey": self.sort_order.value,
            "poll": "false" if not self.already_seen_message_ids else "true",
            "poll_ids": ",".join(str(x) for x in self.already_seen_message_ids),
            "layout": "new",
        }

    @property
    def _xpath(self) -> str:
        return ".//messages/message"

    @property
    def _object_to_instantiate(self) -> type[ShortMessage]:
        return ShortMessage


class _FetchOneMessage(_MessagesPoster, SmartschoolXML, ABC):
    def __init__(self, msg_id: int, box_type: BoxType = BoxType.INBOX):
        super().__init__()

        self.msg_id = msg_id
        self.box_type = box_type

    @property
    def _subsystem(self) -> str:
        return "postboxes"

    @property
    def _params(self) -> dict:
        return {
            "msgID": self.msg_id,
            "boxType": self.box_type.value,
            "limitList": "true",
        }

    def _get_from_cache(self) -> object:
        return self.cache[(self.msg_id, self.box_type)]

    def _store_into_cache(self, obj: object) -> None:
        self.cache[(self.msg_id, self.box_type)] = obj


class Message(_FetchOneMessage):
    """
    Interface to fetch one message based on its MessageID.

    To reproduce: just go to a message and click it.

    Example:
    -------
    >>> message: FullMessage = list(Message(123))[0]
    >>> message.subject
    Griezelfestijn

    """

    @property
    def _action(self) -> str:
        return "show message"

    @property
    def _xpath(self) -> str:
        return ".//data/message"

    @property
    def _object_to_instantiate(self) -> type[FullMessage]:
        return FullMessage

    def _post_process_element(self, element: dict) -> None:
        for modify_this in ["receivers", "ccreceivers", "bccreceivers"]:
            if not element[modify_this]:
                element[modify_this] = []
                continue

            element[modify_this] = element[modify_this]["to"]

            if not isinstance(element[modify_this], list):
                element[modify_this] = [element[modify_this]]


class Attachments(_FetchOneMessage):
    """
    Interface to fetch one message based on its MessageID.

    To reproduce: just go to a message (with the attachment symbol) and click it.

    Example:
    -------
    >>> for attachment in Attachments(123):
    >>>     print(attachment.name)
    Oproep figuranten en helpers.pdf
    Poster griezelfestijn.pdf

    """

    @property
    def _action(self) -> str:
        return "attachment list"

    @property
    def _xpath(self) -> str:
        return ".//attachmentlist/attachment"

    @property
    def _object_to_instantiate(self) -> type[Attachment]:
        return Attachment


class MarkMessageUnread(_FetchOneMessage):
    @property
    def _action(self) -> str:
        return "mark message unread"

    @property
    def _xpath(self) -> str:
        return ".//data/message"

    @property
    def _object_to_instantiate(self) -> type[MessageChanged]:
        return MessageChanged

    @property
    def _params(self) -> dict:
        return {
            "boxType": self.box_type.value,
            "boxID": "0",
            "msgID": self.msg_id,
            "clAction": "status",
        }


class AdjustMessageLabel(_FetchOneMessage):
    def __init__(self, msg_id: int, box_type: BoxType = BoxType.INBOX, label: MessageLabel = MessageLabel.NO_FLAG):
        super().__init__(msg_id, box_type)
        self.label = label

    @property
    def _action(self) -> str:
        return "save msglabel"

    @property
    def _xpath(self) -> str:
        return ".//data/message"

    @property
    def _object_to_instantiate(self) -> type[MessageChanged]:
        return MessageChanged

    @property
    def _params(self) -> dict:
        return {
            "boxType": self.box_type.value,
            "msgLabel": self.label.value,
            "msgID": self.msg_id,
            "clAction": "label",
        }


class MessageMoveToArchive:
    """
    Archiving is weird.

    It's not following the XML protocol... Providing the same interface as the other XMLs though.
    """

    def __init__(self, msg_id: int | list[int]):
        super().__init__()

        if not isinstance(msg_id, list):
            msg_id = [msg_id]

        self.msg_ids = msg_id

    def get(self) -> MessageChanged:
        return next(iter(self))

    def __iter__(self) -> Iterator[MessageChanged]:
        construction = "&".join("msgIDs%5B%5D=" + quote_plus(str(msg_id)) for msg_id in self.msg_ids)

        resp = session.post(
            "/Messages/Xhr/archivemessages",
            data=construction,
            headers={
                "X-Requested-With": "XMLHttpRequest",
            },
        )

        success = resp.json()["success"]
        for msg_id in self.msg_ids:
            yield MessageChanged(id=msg_id, new=1 if msg_id in success else 0)


class MessageMoveToTrash(_MessagesPoster, SmartschoolXML_NoCache):
    def __init__(self, msg_id: int):
        super().__init__()

        self.msg_id = msg_id

    @property
    def _subsystem(self) -> str:
        return "postboxes"

    @property
    def _action(self) -> str:
        return "quick delete"

    @property
    def _xpath(self) -> str:
        return ".//data/details"

    @property
    def _object_to_instantiate(self) -> type[MessageDeletionStatus]:
        return MessageDeletionStatus

    @property
    def _params(self) -> dict:
        return {"msgID": self.msg_id}
